/******************************************************************************/
/*   IBM Confidential                                                         */
/*                                                                            */
/*   OCO Source Materials.                                                    */
/*                                                                            */
/*   (C) Copyright IBM Corp. 1999, 2000, 2001, 2002.                                */
/*                                                                            */
/*   The source code for this program is not published or otherwise           */
/*   divested of its trade secrets, irrespective of what has been             */
/*   deposited with the U.S. Copyright Office.                                */
/*                                                                            */
/*   Product ID:                                                              */
/******************************************************************************/

/**
 * Filename: isyncsample.c
 * 
 * Description:  
 *
 *      This is an example of how to use the IBM Sync API.  
 *
 */

/*
 * To use the IBM Sync API, the user simply needs to do the following: 
 *
 *     1) Include system-specific standard header files
 *
 *     2) Include isyncore.h C header file on your application.
 *
 *     3) Use the IBM Sync Client API in your application by following the specification
 *
 *     4) Compile and link your application object code with the IBM Sync static libraries
 *        "isyncconf.lib" and "isyncore.lib"
 *
 *     5) Setup the Path for the IBM Sync engine.
 *        set PATH to which the IBM Sync engine libraries reside, so
 *        the operating system can locate it.
 *
 *        The IBM Sync engine libraries consist of:
 *
 *         wbxmllib
 *         imsaconfig
 *         imsadb2e
 *         imsafile
 *         isyncconf
 *         isyncore
 */

// 1) system-specific header files
#include <stdio.h>
#include <stdlib.h>

// 2) Include isyncore.h C header file on your application
#include "isyncore.h"

// utility functions
static void printApps(HISCCONF hConfig);

// isync callback prototype
static int syncListener(isy_UINT32 listenerData, ISCEVT *event, isy_VOID *pExtraInfo);

/******************************************************************/
// start of main
int main(int argc, char *argv[])
{
  HISCSERV  hService = NULL;		
  HISCCONF  hConfig = NULL;
  HISCENG   hEngine = NULL;
  isy_INT32 rc;

  printf("Start of ISync Sample...\n");

  /**
   *  (1) Open required handles
   */
  // Open a service handle
  rc = iscServiceOpen("127.0.0.1", "8080", "db2admin", "db2admin", NULL, &hService);
  // Open a config handle
  rc = iscConfigOpen(hService, ".\\", &hConfig);
  // Open a handle to the sync engine
  rc = iscEngineOpen(hConfig, &hConfig);

  /**
   *  (2) Set the preferences of the IBM Sync engine
   */
  iscEngineSetPref(hEngine, ISCPREF_Trace,   ISCCONST_TraceON); // Turn on trace
  iscEngineSetPref(hEngine, ISCPREF_Timeout, "120");            // 2 min transport timeout

  /**
   *  (3) Register the callback to the IBM Sync engine for event notification if needed
   *      by the application.  The syncListener() is defined below.
   */
  iscEngineSetListener(hEngine, syncListener, (isy_UINT32) NULL);

  /**
   *  (4) Perform the synchronization.
   */
  iscEngineSync(hEngine);

  /**
   *  (5) Print the sync status of all the subscription sets.
   */
  printApps(hConfig);

  /**
   *  (6) Synchronization is done, close all handles
   */
  iscEngineClose(hEngine);
  iscConfigClose(hEngine);
  iscServiceClose(hEngine);

  printf("End of ISync Sample.\n");

  return 0;
}

/******************************************************************/
/**
 * This function displays the applications to be sync'd on the client.
 * param: hEngine, HISYNC instance used to call IBM Sync API.
 * return: nothing
 **/
static void printApps(HISCCONF hConfig)
{
  HISCCSR   hCursor;
  char      id[ISCLEN_SubsSetID], name[ISCLEN_SubsSetName];
  isy_INT32 rc;
  int       i;

  printf("*******APPINFO*******\n");

  rc = iscConfigOpenCursor(hConfig, &hCursor);
  if (rc != ISCRTN_Succeeded) return;

  i = 0;
  while (iscConfigGetNextSubsSet(hConfig, hCursor, id, name) == ISCRTN_Succeeded)
  {
    rc = iscConfigGetSubsSetStatus(hConfig, id);

    i++;
    printf("SubsSet #%i: %s (id: %s)", i, name, id);
	if (rc == ISCRTN_Succeeded)
      printf("OK\n");
	else if (rc == ISCRTN_Canceled)
      printf("Canceled\n");
	else
      printf("Failed\n");
  }
  if (i == 0)
    printf("There is no subscription set.\n");
  printf("*****APPINFO END*****\n");
  iscConfigCloseCursor(hConfig, hCursor);
}

/******************************************************************/
/**
 * This function defines the sync listener. See isyncore.h for more
 * information.
 * param: listenerData, your personal data.
 * param: event, event object
 * param: pExtraInfo (reserved)
 * return: integer, when event type is ISCEVTTYPE_Retry:
 *                  . ISCRTNCB_ReplyYes    : retry less than 3 times
 *                  . ISCRTNCB_ReplyNo     : retry more than or equal to 3 times
 *                  when event type is ISCEVTTYPE_Info:
 *                  . ISCRTNCB_Done
 *                  others (ISCEVTTYPE_Fatal, ISCEVTTYPE_Error, ISCEVTTYPE_Query,
 *                          and ISCEVTTYPE_Conflict)
 *                  . ISCRTNCB_Default     : take default action
 **/
static int syncListener(
       isy_UINT32 listenerData,
       ISCEVT    *event,
       isy_VOID  *pExtraInfo)
{
  int   timesRetried;

  switch (event->type) {
    case ISCEVTTYPE_Fatal:
    case ISCEVTTYPE_Error:
      printf("Error: %d\n", event->code);
      return ISCRTNCB_Default ;

    case ISCEVTTYPE_Retry:
      timesRetried = event->retry;
      if (timesRetried >= 3)
        return ISCRTNCB_ReplyNo;
      else {
        char ans;
        printf("Retry? [Y/N]: ");
        ans = getchar();        
        getchar();
        if (tolower(ans) == 'y')
          return ISCRTNCB_ReplyYes;
        else
          return ISCRTNCB_ReplyNo;
      }

    case ISCEVTTYPE_Info:
      switch (event->code) {
        case ISCEVT_InfSyncSucceeded:
          printf("Conclusion: synchronization succeeded\n");
          break;
        case ISCEVT_InfSyncFailed:
          printf("Conclusion: synchronization failed\n");
          break;
        case ISCEVT_InfSyncCanceled:
          printf("Conclusion: synchronization canceled\n");
          break;
        default: // ignore other event code
          printf("Status(%d%%): %d\n", event->state.syncProg, event->code);
          break;
      } // switch (event->code)
      return ISCRTNCB_Done;

    // all other event types, don't care, simply take the default action
    default:
      return ISCRTNCB_Default;
  } // switch (event->type)
}
	
/********************************************************************/


