/************************************************************************
 *
 * Licensed Materials - Property of IBM
 * Program Name: IBM DB2Everyplace Sync Client GUI
 * (c ) Copyright IBM Corp. 1999-2003. All Rights Reserved.
 *
 * US Government Users Restricted Rights 
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. Licensed Materials - Property of IBM
 *
 * IBM grants you ("Licensee") a non-exclusive, royalty free, license to
 * use Sample Code.  This software is provided as-is and without any
 * warranties, whether EXPRESS OR  IMPLIED, INCLUDING ANY IMPLIED
 * WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR 
 * NON-INFRINGEMENT. IBM AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY 
 * DAMAGES SUFFERED BY LICENSEE THAT RESULT FROM YOUR USE OF THE SOFTWARE.
 * IN NO EVENT WILL IBM OR ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE,
 * PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, 
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE 
 * THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE 
 * SOFTWARE, EVEN IF IBM HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES.
 *
 ************************************************************************
 */
/************************************************************************
 *
 * File Name:  db2sync_console.c
 * Description:
 *    An interactive example application using the ISync API.  Most of
 * the features of the ISync GUI applications (ie. Palm) are present in
 * this example.  These features include enabling, disabling or resetting 
 * each subscription set, editting the user and network settings that 
 * enable synchronizations with the server, viewing the log, and of 
 * course starting the synchronization.  As a part of this ISync API 
 * example, there is an example of a syncListener which enables callbacks
 * from the API.
 *
 * The customer can develop your own IBM DB2Everyplace Sync Client by
 * following the following steps:
 *
 *     1) Include system-specific standard header files in the source
 *
 *     2) Include isyncore.h C header file on your application.
 *
 *     3) Use the IBM SyncClient API according to the specification in 
 *        the Application Development Guide
 *
 *     4) Compile and link your application object code with two static
 *        libraries, isyncconf and isyncore
 *
 *     5) Setup the Path for the IBM Sync engine.
 *        set PATH to which the IBM Sync engine libraries reside, so the
 *        operating system can locate them.
 *
 *        The IBM Sync engine libraries consist of:
 *
 *        isyncconf  : Configuration management module
 *        isyncore   : IBM Sync engine
 *        isyncxpt   : Message transportation module
 *        isynce     : IBM Sync V7.2 API compability library
 *        imsaconfig : Configuration adapter
 *        imsadb2e   : DB2e adapter
 *        imsafile   : File adapter
 *        wbxmllib   : WBXML library
 *        SSLLib     : SSL library (PalmOS only)
 *
 ************************************************************************
 */

// 1) system-specific header files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// 2) Include isyncore.h C header file on your application
#include "isyncore.h"

#include "messages.h"
#include "db2sync_console.h"
#include "osdefs.h"

#ifdef UNDER_LN
#include "locale.h"	
#endif

typedef struct
{
	isy_UINT32 reserved1;
	isy_UINT16 reserved2;
	isy_BYTE   reserved3[24];
	isy_BYTE   options; // trace

	isy_TCHAR  reserved5[20];
	long       reserved6;
	long       reserved7;
	long       reserved8;
	long       timeout;
	isy_TCHAR  reserved10[56];
	isy_TCHAR  user[256];		
	isy_TCHAR  password[56];	
	isy_TCHAR  reserved11[80];	
	isy_TCHAR  dest[50];
	isy_TCHAR  reserved13[220];	
	isy_TCHAR  path[100];					
	isy_INT32  messageSize;
	isy_TCHAR  proxydest[100];
	isy_TCHAR  URL[513];
}Prefs;

//Global variables for preferences and handles
HISCSERV  hSyncServ = NULL;		
HISCCONF  hSyncConf = NULL;
HISCENG   hSyncer = NULL;
Prefs     myPref; 
FILE     *hLog = NULL;
isy_TCHAR subsTarget[256];

//Main functions
static int  initSettings();     //default settings
static void goSync();			//begins synchronization
static void goSettings();		//edits settings
static void goLog();			//prints the log
static void goApps();			//alters an applications mode
static void goAbout();			//about SyncClient
static void goExit();			//quits this program
static void openLog();
static void rewindLog();
static void closeLog();
static int  syncStartup();
static void syncShutdown();
static void logEvent(FILE *hLog, ISCEVT*  event, isy_TCHAR* msg);
static void logWrite(FILE *hLog, isy_TCHAR *buf);
static void logCodePage(FILE* hLog);
const isy_TCHAR* appCurrentDateTime(isy_TCHAR* buf, isy_TCHAR* formatString);


//
//Callback function for ISync API
static isy_INT32  syncListener(
       isy_UINT32 listenerData,
       ISCEVT    *event,
       isy_VOID  *pExtraInfo);

//
//Helper functions
static isy_TCHAR getAction();

static int getPrefs(Prefs* pPrefs);
static void setPrefs(Prefs* pPrefs);

isy_TCHAR* appEventToMessage(ISCEVT*  event);

/******************************************************************/
/** 
 * This is where it all starts. No arguments are used.
 * This function loops until getAction switches to DO_EXIT and the
 * program is terminated.
 * Params: None
 * Returns: Never reached
 **/

int main(int argc, isy_TCHAR *argv[])
{
	int rc;
	isy_TCHAR   action;		
	
#ifdef UNDER_LN
	setlocale(LC_CTYPE, "");
#endif
	rc = initSettings();
	
	PRINTF (isy_T("\n\n"));

	while(1)
	{
		action = getAction();
		switch (action){
		case DO_SYNC:  
			PRINTF (isy_T("\n\n"));			
			goSync();			
			break;
		case DO_APPS:  
			PRINTF (isy_T("\n\n"));			
			goApps();			
			break;			
		case DO_PREFS: 
			PRINTF (isy_T("\n\n"));			
			goSettings();			
			break;
		case DO_LOG:   
			PRINTF (isy_T("\n\n"));
			goLog();
			break;
		case DO_ABOUT:  
			goAbout();
			break;
		case DO_EXIT:  
			goExit();
			break;
		case DO_NOTHING:  
			break;
		default:  
			PRINTF(isy_T("'%c' %s\n\n"), action, isy_T(DB2SYNC_InvalidChoice));
			break;
		}
	}
	
	return 0;
}

/******************************************************************/
/**
 * Prompts the user for an action to perform and returns an int
 * representing their choice.
 * Params: None
 * Returns: int, the users choice
 **/
static isy_TCHAR getAction()
{
	isy_TCHAR ch, action = DO_NOTHING;

	PRINTF(isy_T(DB2SYNC_MainMenu));
	PRINTF(isy_T(DB2SYNC_MainMenu1));
	PRINTF(isy_T(DB2SYNC_MainMenu2));
	PRINTF(isy_T(DB2SYNC_MainMenu3));
	PRINTF(isy_T(DB2SYNC_MainMenu4));
	PRINTF(isy_T(DB2SYNC_MainMenu5));
	PRINTF(isy_T(DB2SYNC_MainMenu6));
	PRINTF(isy_T(DB2SYNC_MakeChoice));
	while((ch = getchar()) != '\n')
	{
		if (action == DO_NOTHING)
			action = ch;
	} 
	return (int)action;
}

/******************************************************************/
/**
 * Performs the synchronization.  Before isyncGo is called, the
 * syncListener is set so that the API can perform callbacks.
 * Params: None
 * Returns: None
 **/
static void goSync()
{	
	int rc;
	
	rc = syncStartup(); 
	if (rc != ISCRTN_Succeeded)
	{
		PRINTF(isy_T("\n%s %d\n\n\n"), ISYNCSTR_StartupError, rc);
		return;
	}
	rewindLog();

	iscEngineSetListener(hSyncer, syncListener, (isy_UINT32) NULL);
	
	rc = iscEngineSync(hSyncer);
	PRINTF(isy_T("iscEngineSync() = %d\n\n"), rc);
	
	{
        HISCCSR    hCursor;
        isy_TCHAR  id[ISCLEN_SubsSetID];
        isy_TCHAR  name[ISCLEN_SubsSetName];
        isy_INT32  enabled, reset;

        iscConfigOpenCursor(hSyncConf, &hCursor);
		PRINTF(isy_T("%s"), isy_T(DB2SYNC_SyncStatus));
        while (iscConfigGetNextSubsSet(hSyncConf, hCursor, id, name)
               == ISCRTN_Succeeded) {
            enabled = iscConfigSubsSetIsEnabled(hSyncConf, id);
            reset = iscConfigSubsSetIsReset(hSyncConf, id);
            rc = iscConfigGetSubsSetStatus(hSyncConf, id);
            if (rc == ISCRTN_Succeeded) {
				PRINTF(isy_T("%c%8s %28s %s\n"), (enabled == ISCRTN_True ? '+' : '-'), ISYNCSTR_Succeeded, id, name);
			} else if (rc == ISCRTN_Failed) {
				PRINTF(isy_T("%c%8s %28s %s\n"), (enabled == ISCRTN_True ? '+' : '-'), ISYNCSTR_Failed, id, name);
			} else if (rc == ISCRTN_Canceled) {
				PRINTF(isy_T("%c%8s %28s %s\n"), (enabled == ISCRTN_True ? '+' : '-'), ISYNCSTR_Cancelled, id, name);
			} else if (rc == ISCRTN_Ready) {
				PRINTF(isy_T("%c%8s %28s %s\n"), (enabled == ISCRTN_True ? '+' : '-'), ISYNCSTR_Ready, id, name);
			} else {
				PRINTF(isy_T("%c%8s %28s %s\n"), (enabled == ISCRTN_True ? '+' : '-'), isy_T("?"), id, name);
			}
        }
		PRINTF(isy_T("%s"), isy_T(DB2SYNC_Line));
        iscConfigCloseCursor(hSyncConf, hCursor);
    }
	
	PRINTF (isy_T("\n\n"));

	closeLog();
	syncShutdown();
}

/******************************************************************/
/**
 * Allows the user to modify the state of a subscription set.  
 * Each subscription set can be enabled, disabled, or reset.
 * If no changes are necessary to the sync application the user
 * can press enter to accept the default setting.  
 * Params: None
 * Returns: None
 **/
static void goApps()
{
	HISCCSR    hCursor;
	isy_TCHAR  id[ISCLEN_SubsSetID];
	isy_TCHAR  name[ISCLEN_SubsSetName];
	isy_INT32  enabled, reset;
	int        haveApps = 0, done = 0, rc;

	rc = syncStartup(); 
	if (rc != ISCRTN_Succeeded)
	{
		PRINTF(isy_T("\n%s %d\n\n\n"), ISYNCSTR_StartupError, rc);
		return;
	}

	PRINTF(isy_T(DB2SYNC_SyncApps));

	iscConfigOpenCursor(hSyncConf, &hCursor);
	rc = iscConfigGetNextSubsSet(hSyncConf, hCursor, id, name);
	while (!done && rc == ISCRTN_Succeeded) 
	{
		isy_TCHAR ch, mode = MODE_NONE;
		int       modeSet = 0;

		haveApps = 1;
		enabled = iscConfigSubsSetIsEnabled(hSyncConf, id);
		reset = iscConfigSubsSetIsReset(hSyncConf, id);

		PRINTF(isy_T(DB2SYNC_AppMode), name,
			(enabled ? isy_T(DB2SYNC_Enabled) : isy_T(DB2SYNC_Disabled)),
			(reset ? isy_T(DB2SYNC_ModeRefresh) : isy_T(DB2SYNC_ModeSynchronize)));

		PRINTF(isy_T(DB2SYNC_SetAppMode));
		PRINTF(isy_T(DB2SYNC_SetAppModeA));
		PRINTF(isy_T(DB2SYNC_SetAppModeB));		
		PRINTF(isy_T(DB2SYNC_SetAppModeC));
		PRINTF(isy_T(DB2SYNC_SetAppModeD));
		PRINTF(isy_T(DB2SYNC_MakeChoice));
		
		// Get the mode
		while((ch = getchar()) != '\n')
			if (mode == MODE_NONE) mode = ch;

		switch(mode)
		{
		case MODE_ENABLE:
			iscConfigEnableSubsSet(hSyncConf, id);
			modeSet = 1;				
			break;
		case MODE_DISABLE:
			iscConfigDisableSubsSet(hSyncConf, id);				
			modeSet = 1;				
			break;
		case MODE_RESET:
			iscConfigResetSubsSet(hSyncConf, id);				
			modeSet = 1;				
			break;
		case MODE_DONE:
			done = 1;				
			modeSet = 1;				
			break;
		case MODE_NONE:
			modeSet = 1;
			break; // ignore the return key
		default:
			modeSet = 0; // Invalid choice, try again.	
			PRINTF(isy_T("'%c' %s\n\n"), mode, isy_T(DB2SYNC_InvalidChoice));
			break;
		}												
		PRINTF (isy_T("\n\n"));
		if (modeSet) {
		    rc = iscConfigGetNextSubsSet(hSyncConf, hCursor, id, name);
		}
	}
	
	
	if (haveApps == 0) 
	{
		PRINTF(isy_T(DB2SYNC_NoApps));
		return;
	}    

	syncShutdown();
}

/******************************************************************/
/**
 * Do default initial settings, or read previously saved settings if any
 * Params: None
 * Returns: handle to ISYNC
 **/
static int initSettings()
{
	int rc = 0;

	myPref.messageSize = DEFAULT_MSG_SIZE;

	if(getPrefs(&myPref) == -1)
	{
		TSTR_CPY(myPref.dest, isy_T(DEFAULT_DEST));
		TSTR_CPY(myPref.URL,  isy_T(DEFAULT_DEST));
		TSTR_CPY(myPref.path, isy_T(DEFAULT_PATH));
		TSTR_CPY(myPref.user, isy_T(DEFAULT_USER));
		TSTR_CPY(myPref.password, isy_T(DEFAULT_PASSWORD));
		if(TSTR_CMP(isy_T(DEFAULT_TRACE), isy_T(DB2SYNC_TraceDetailed)) == 0)
			myPref.options |= ISYNCOPTION_Trace;
		else
			myPref.options &= ~ISYNCOPTION_Trace;

		myPref.timeout = DEFAULT_TIMEOUT;	
		myPref.messageSize = DEFAULT_MSG_SIZE;
		
		TSTR_CPY(myPref.proxydest, isy_T(DB2SYNC_ProxyFormat));
		
		setPrefs(&myPref);
	}
		
	return rc;
}// initSettings

/******************************************************************/
/**
 * Allows the user to change the network and user settings which
 * enable the sync connection to the server. If no changes are 
 * necessary, the user can press enter to accept the current setting.
 * Params: None
 * Returns: None
 **/

#define	NUM_NETTYPEBAUD	10
typedef struct
{
	isy_TCHAR *prompt;
	isy_INT32  baudRate;
} NetTypeBaud;

static NetTypeBaud NetBaudList[NUM_NETTYPEBAUD] =
	{{isy_T("0: Default"),   0},
	 {isy_T("1: 4K Modem"),  (4   * 1024) / 4},
	 {isy_T("2: 9K GSM"),    (9   * 1024) / 4},
	 {isy_T("3: 19K Modem"), (19  * 1024) / 4},
	 {isy_T("4: 56K Modem"), (56  * 1024) / 4},
	 {isy_T("5: 64K GPRS"),  (64  * 1024) / 4},
	 {isy_T("6: 112K ISDN"), (112 * 1024) / 4},
	 {isy_T("7: 256K DSL"),  (256 * 1024) / 4},
	 {isy_T("8: 2M 	MTS"),   (2   * 1024 * 1024) / 4},
	 {isy_T("9: 10M T1"),    (10  * 1024 * 1024) / 4}};

static void goSettings()
{	
	isy_TCHAR ch;
	isy_TCHAR buf[1024];
	int  i, validTimeout, validNetType;       

	PRINTF(isy_T(DB2SYNC_Settings));
	PRINTF(isy_T(DB2SYNC_SetSettings)); 

	//Get Server URL
	PRINTF(isy_T("%s [%s]: "), isy_T(DB2SYNC_URL), myPref.URL);
	if((ch = getchar()) != '\n')
	{
		for (;ch == ' ';) ch = getchar(); // trim the leading spaces
		for(i = 0; ch != '\n'; i++)
		{	
		  buf[i] = ch;
		  ch = getchar();
		}
		do buf[i--] = '\0'; while (i >= 0 && buf[i] == ' '); // trim the trailing spaces
		if (i > 0) {
			TSTR_CPY(myPref.URL, buf);
			TSTR_N_CPY(myPref.dest, buf, 50);
			myPref.dest[49] = (isy_TCHAR) '\0';
		}
	}

	//Ask User if they want to use proxy	
	PRINTF(isy_T("%s (%s/%s)? "), isy_T(DB2SYNC_UseProxy), isy_T(DB2SYNC_Yes), isy_T(DB2SYNC_No));
	
	if(myPref.options & ISYNCOPTION_Proxy)
		PRINTF(isy_T("[%s]: "), isy_T(DB2SYNC_Yes));
	else
		PRINTF(isy_T("[%s]: "), isy_T(DB2SYNC_No));

	if((ch = getchar()) != '\n')
	{
		for (;ch == ' ';) ch = getchar(); // trim the leading spaces
		for(; getchar() != '\n';);
		if (ch == *DB2SYNC_No || toupper(ch) == *DB2SYNC_No) {
			myPref.options &= ~ISYNCOPTION_Proxy;			
		} else if (ch == *DB2SYNC_Yes || toupper(ch) == *DB2SYNC_Yes) {
			myPref.options |= ISYNCOPTION_Proxy;
		}
	}

	if(myPref.options & ISYNCOPTION_Proxy)
	{
		//Get Proxy Server and port
		PRINTF(isy_T("%s [%s]: "), isy_T(DB2SYNC_Proxy), myPref.proxydest);
		if((ch = getchar()) != '\n')
		{
			for (;ch == ' ';) 
				ch = getchar(); // trim the leading spaces

			for(i = 0; ch != '\n'; i++)
			{	
			  buf[i] = ch;
			  ch = getchar();
			}
			do buf[i--] = '\0'; while (i >= 0 && buf[i] == ' '); // trim the trailing spaces

			if (i >= 0) 
				TSTR_CPY(myPref.proxydest, buf);		
		}
	}
	
	//Get Username	
	PRINTF(isy_T("%s [%s]: "), isy_T(DB2SYNC_User), myPref.user);
	if((ch = getchar()) != '\n')
	{
		for (;ch == ' ';) ch = getchar(); // trim the leading spaces
		for(i = 0; ch != '\n'; i++)
		{	
		  buf[i] = ch;
		  ch = getchar();
		}
		do buf[i--] = '\0'; while (i >= 0 && buf[i] == ' '); // trim the trailing spaces
		if (i >= 0) TSTR_CPY(myPref.user, buf);		
	}
	
	//Get Password
	PRINTF(isy_T("%s ["), isy_T(DB2SYNC_Password));
	for (i = TSTR_LEN(myPref.password); i > 0; i--) PRINTF(isy_T("*"));
	PRINTF(isy_T("]: "));
	i = 0; // no password change
	if((ch = getchar()) != '\n')
	{
		for(i = 0; ch != '\n'; i++)
		{	
		  buf[i] = ch;
		  ch = getchar();
		}		
		do buf[i--] = '\0'; while (i >= 0 && buf[i] == ' '); // trim the trailing spaces
		if (i >= 0) TSTR_CPY(myPref.password, buf);
	}
	
	//Get Path
	PRINTF(isy_T("%s [%s]: "), isy_T(DB2SYNC_Path), myPref.path);
	if((ch = getchar()) != '\n')
	{
		for (;ch == ' ';) ch = getchar(); // trim the leading spaces
		for(i = 0; ch != '\n'; i++)
		{	
		  buf[i] = ch;
		  ch = getchar();
		}
		do buf[i--] = '\0'; while (i >= 0 && buf[i] == ' '); // trim the trailing spaces
		if (i >= 0) TSTR_CPY(myPref.path, buf);		
	}

	//Get Trace ON/OFF	
	PRINTF(isy_T("%s (0:%s / 1:%s)? "), isy_T(DB2SYNC_Trace), isy_T(DB2SYNC_TraceDefault), isy_T(DB2SYNC_TraceDetailed));
	
	if(myPref.options & ISYNCOPTION_Trace)
		PRINTF(isy_T("[%s]: "), isy_T("1"));
	else
		PRINTF(isy_T("[%s]: "), isy_T("0"));

	if((ch = getchar()) != '\n')
	{
		for (;ch == ' ';) ch = getchar(); // trim the leading spaces
		for(; getchar() != '\n';);
		if (ch == '0') {
			myPref.options &= ~ISYNCOPTION_Trace;			
		} else if (ch == '1') {
			myPref.options |= ISYNCOPTION_Trace;
		}
	}

	//Get Timeout length	
	//validTimeout = TSTR_ATOI(myPref.timeoutStr);
	validTimeout = myPref.timeout;
	if (validTimeout == 0) {
		PRINTF(isy_T("%s [%s]: "), isy_T(DB2SYNC_Timeout), isy_T(DB2SYNC_MinimumTimeout));
	} else if (validTimeout == -1) {
		PRINTF(isy_T("%s [%s]: "), isy_T(DB2SYNC_Timeout), isy_T(DB2SYNC_NeverTimeout));
	} else {
		validTimeout /= 60; // minutes
		TSTR_ITOA(buf, validTimeout);
		PRINTF(isy_T("%s [%s min]: "), isy_T(DB2SYNC_Timeout), buf);
	}
	if((ch = getchar()) != '\n')
	{
		for (;ch == ' ';) ch = getchar(); // trim the leading spaces
		validTimeout = 1;
		for(i = 0; ch != '\n'; i++)
		{	
		  buf[i] = ch;
		  if (ch != '-' && (ch < '0' || ch > '9')) validTimeout = 0;
		  ch = getchar();
		}
		do buf[i--] = '\0'; while (i >= 0 && buf[i] == ' '); // trim the trailing spaces
		if (validTimeout && i >= 0) {
			validTimeout = TSTR_ATOI(buf);
			if (validTimeout > 0) 
				validTimeout *= 60; // seconds
			myPref.timeout = validTimeout;
			
			//TSTR_ITOA(buf, validTimeout);
			//TSTR_CPY(myPref.timeoutStr, buf);			
		} else
			PRINTF(isy_T("%s\n"), isy_T(DB2SYNC_InvalidTimeout));
	}
	

	TSTR_CPY(buf, isy_T("0"));
	for (i = 0; i < NUM_NETTYPEBAUD; i++)
	{
		PRINTF(isy_T("%s=%d\n"),
		NetBaudList[i].prompt, NetBaudList[i].baudRate);
		if (myPref.messageSize == NetBaudList[i].baudRate) TSTR_ITOA(buf, i);
	}
	PRINTF(isy_T("%s [%s]: "), isy_T(DB2SYNC_NetworkType), buf);

	if((ch = getchar()) != '\n')
	{
		validNetType = 1;
		for(i = 0; ch != '\n'; i++)
		{	
		  buf[i] = ch;
		  buf[i+1] = '\0';
		  if (ch == '-' || (ch < '0' || ch > '9')) validNetType = 0;
		  ch = getchar();
		}
		if (validNetType)
		{
			int index = TSTR_ATOI(buf);
			myPref.messageSize = NetBaudList[index].baudRate;
		}
		else
			PRINTF(isy_T("%s\n"), isy_T(DB2SYNC_InvalidNetworkType));
	}
	// PRINTF("Picked baud rate %d\n", myPref.messageSize);
	PRINTF (isy_T("\n\n"));
	
	setPrefs(&myPref);
}

/******************************************************************/
/**
 * Prints the log file to the screen.
 * Params: None
 * Returns: None
 **/
static void goLog()
{
	int   ch;

	openLog();

	if (hLog == NULL)
	{
       return;
	}

	PRINTF(isy_T(DB2SYNC_LogFile));
	while((ch = getc(hLog)) != EOF)
	{
		PRINTF(isy_T("%c"), ch);
	}
	PRINTF (isy_T("\n\n"));

	closeLog();
}

/******************************************************************/
/**
 * Releases all the handle
 * Params: None
 * Returns: None
 **/
static void syncShutdown()
{
	if (hSyncer != NULL) {
		iscEngineClose(hSyncer);
		hSyncer = NULL;
	}

	if (hSyncConf != NULL) {
		iscConfigClose(hSyncConf);
		hSyncConf = NULL;
	}

	if (hSyncServ != NULL) {
		iscServiceClose(hSyncServ);
		hSyncServ = NULL;
	}		
}

/******************************************************************/
/**
 * Initializes all the handle
 * Params: None
 * Returns: None
 **/
static int syncStartup() 
{
	ISCPROPERTY properties[16] = {{ISCPROP_SyncUser,        NULL},
	                              {ISCPROP_SyncPassword,    NULL},
	                              {ISCPROP_SyncMessageSize, NULL},
	                              {ISCPROP_SyncTrace,       NULL},
	                              {ISCPROP_SyncTimeout,     NULL},								
	                              {ISCPROP_SyncProxy,       NULL}};
	int       numProperty = 0;
	isy_TCHAR messageSize[20];
	isy_TCHAR timeout[20];
	int       rc = 0;

	syncShutdown();   

	TSTR_ITOA(messageSize, myPref.messageSize);
	TSTR_ITOA(timeout, myPref.timeout);
	properties[numProperty++].value = myPref.user;			
	properties[numProperty++].value = myPref.password;
	properties[numProperty++].value = messageSize;		
	properties[numProperty++].value = (myPref.options & ISYNCOPTION_Trace) ? isy_T("detailed") : isy_T("default");
	properties[numProperty++].value = timeout;	

	if (myPref.options & ISYNCOPTION_Proxy)
		properties[numProperty++].value = myPref.proxydest;

	// rc = iscServiceOpen(server, port, myPref.user, myPref.password, NULL, &hSyncServ);
	rc = iscServiceOpenEx(myPref.URL, properties, numProperty, &hSyncServ);
	if (rc != ISCRTN_Succeeded) return rc;

	rc = iscConfigOpen(hSyncServ, myPref.path, &hSyncConf);
	if (rc != ISCRTN_Succeeded) return rc;

	rc = iscEngineOpen(hSyncConf, &hSyncer);

	return rc;
}
/******************************************************************/
/**
 * About the SyncClient
 * Params: None
 * Returns: None
 **/

static void goAbout() 
{
	isy_TCHAR about[256];
	int rc;

	rc = syncStartup(); 
	if (rc != ISCRTN_Succeeded)
	{
		PRINTF(ISYNCSTR_StartupError);
		PRINTF(isy_T("%d\n\n\n"), rc);
		return;
	}
	iscEngineGetInfo(hSyncer, about, 256);
	PRINTF(isy_T("\n%s\n%s\n\n"), isy_T(DB2SYNC_About), about);
	syncShutdown();
}

/******************************************************************/
/**
 * Exits this program.
 * Params: None
 * Returns: None
 **/

static void goExit() 
{
	PRINTF(isy_T("%s ...\n"), isy_T(DB2SYNC_Exitting));
	syncShutdown();
	exit(0);
}

#define eventTypeSymbol(event)	\
				((event->type == ISCEVTTYPE_Info)     ? 'I' : \
				 (event->type == ISCEVTTYPE_Retry)    ? 'R' : \
				 (event->type == ISCEVTTYPE_Query)    ? 'Q' : \
				 (event->type == ISCEVTTYPE_Error)    ? 'E' : \
				 (event->type == ISCEVTTYPE_Fatal)    ? 'F' : \
				 (event->type == ISCEVTTYPE_Conflict) ? 'C' : '?')

#define printEventStateFormat(event)	\
				isy_T("DSYC%04d%c(%d%%): %s"), \
				event->code, \
				eventTypeSymbol(event), \
				event->state.syncProg

#define printEventState(event, statusMsg)	\
				PRINTF(printEventStateFormat(event), statusMsg)

/******************************************************************/
/**
 * This function defines the sync listener. See isyncore.h for more
 * information.
 * param: listenerData, your personal data.
 * param: event, event object
 * param: pExtraInfo (reserved)
 * return: integer, when event type is ISCEVTTYPE_Retry:
 *                  . ISCRTNCB_ReplyYes    : retry less than 3 times
 *                  . ISCRTNCB_ReplyNo     : retry more than or equal to 3 times
 *                  when event type is ISCEVTTYPE_Info or ISCEVTTYPE_Conflict:
 *                  . ISCRTNCB_Done        : don't care
 *                  other event types
 *                  . ISCRTNCB_Default
 **/
static isy_INT32  syncListener(
       isy_UINT32 listenerData,
       ISCEVT    *event,
       isy_VOID  *pExtraInfo)
{
	isy_TCHAR statusMsg[200];
	int       timesRetried;
	int       querySubsTarget = 0; // Take default action for the per-subscription target directory
	isy_INT32 rc = ISCRTNCB_Default;	
	
	statusMsg[0] = '\0';	

	if (event == NULL) {
		PRINTF (isy_T("event is NULL \n"));
		return rc;
	}
	
	if((event->code == ISCEVT_ErrConnectData) && (event->state.subsType == ISCSUBSTYPE_File))
	{
		TSTR_CPY(statusMsg, ISYNCSTR_ErrConnectFile);
	}
	else
		TSTR_CPY(statusMsg, appEventToMessage(event));		

	switch (event->type) {
        	case ISCEVTTYPE_Retry:
			printEventState(event, statusMsg);
			logEvent(hLog, event, statusMsg);
			timesRetried = event->retry;
			if (timesRetried >= 3)
				rc = ISCRTNCB_ReplyNo;
			else {
				isy_TCHAR ans;
				PRINTF(isy_T("\n%s [%s/%s]?"), ISYNCSTR_Retry, isy_T(DB2SYNC_Yes), isy_T(DB2SYNC_No));
						ans = getchar();	
						for(; getchar() != '\n';);
				if(toupper(ans) == *DB2SYNC_Yes)
				        rc = ISCRTNCB_ReplyYes;
				else
				        rc = ISCRTNCB_ReplyNo;
			}
			PRINTF(isy_T("\n"));
			break;
	        case ISCEVTTYPE_Info:			
			printEventState(event, statusMsg);
			switch (event->code) {				
				case ISCEVT_InfSubsSetStarted:
					if (*event->state.currSubsSet)
					{
						PRINTF(isy_T(" \"%s\""), event->state.currSubsSet);
						TSTR_CAT(statusMsg, isy_T(": "));
						TSTR_CAT(statusMsg, event->state.currSubsSet);
						
					}
					logEvent(hLog, event, statusMsg);

					break;
				case ISCEVT_InfSyncingSubs:										
					if (*event->state.currSubs)
					{
						PRINTF(isy_T(" \"%s\""), event->state.currSubs);
						TSTR_CAT(statusMsg, isy_T(": "));
						TSTR_CAT(statusMsg, event->state.currSubs);
					}
					logEvent(hLog, event, statusMsg);


					break;
				default:					
					logEvent(hLog, event, statusMsg);
					break;
			} // switch (event->code)
			PRINTF(isy_T("\n"));
			break;
	        case ISCEVTTYPE_Query:
			if (event->code == ISCEVT_QueCancel)			
				break;			
			else if (event->code == ISCEVT_QueLogin) {
				ISCLISTENARG *args = event->info;
				isy_TCHAR c;
				int  i;

				isy_TCHAR *target = args->argv[0];
				isy_TCHAR *username = args->argv[1];
				isy_TCHAR *password = args->argv[2];
				printEventState(event, statusMsg);
				logEvent(hLog, event, statusMsg);
				PRINTF(isy_T("\"%s\"\n"), target);
				PRINTF(isy_T("%s: "), isy_T(DB2SYNC_User));
				for(i = 0; (c = getchar()) != '\n'; i++) username[i] = c;
				username[i] = '\0';
				if (i > 0) {
					PRINTF(isy_T("%s: "), isy_T(DB2SYNC_Password));
					for(i = 0; (c = getchar()) != '\n'; i++) password[i] = c;
					password[i] = '\0';
					rc = ISCRTNCB_Done;
				}
				break;	
			}

			else if (event->code == ISCEVT_QueSubsTarget && querySubsTarget) {
				ISCLISTENARG *args = event->info;
				ISCSTATE *state = &event->state;
				isy_TCHAR c;
				int  i;

				printEventState(event, statusMsg);
				logEvent(hLog, event, statusMsg);
				// PRINTF(isy_T("\"%s\"\n"), subsTarget);
				PRINTF(isy_T("\n%s: "), isy_T(DB2SYNC_SubsTarget));
				for(i = 0; (c = getchar()) != '\n'; i++) subsTarget[i] = c;
				subsTarget[i] = '\0';
				args->argv[0] = subsTarget;
				if (i > 0)
					rc = ISCRTNCB_Done;
				break;
			}

			break;

		case ISCEVTTYPE_Conflict:
		{
			//from sqlcli.h
			//#define  SQL_INTEGER             4
			//#define  SQL_SMALLINT            5
			#define SQL_BLOB					-98
		
			ISCLISTENCONFLICT *pConflict = (ISCLISTENCONFLICT *) event->info;
			int i=0;
		
			printEventState(event, statusMsg);
			logEvent(hLog, event, statusMsg);
			
			if (!pConflict) 
			{
				PRINTF (isy_T("\n"));
				logWrite(hLog, FILE_NL);
				break;
			}
		
			PRINTF(isy_T("%s %s,"),ISYNCSTR_MSG_CONFLICT,pConflict->table);
		
			switch (pConflict->op)
			{
				case ISCCONST_OpDelete:
					PRINTF(isy_T("%s\n%s"), ISYNCSTR_MSG_OP_DELETE, ISYNCSTR_MSG_DATA);
					logWrite(hLog, ISYNCSTR_MSG_OP_DELETE);
					logWrite(hLog, ISYNCSTR_MSG_DATA);
					break;
				case ISCCONST_OpInsert:
					PRINTF(isy_T("%s\n%s"), ISYNCSTR_MSG_OP_INSERT, ISYNCSTR_MSG_DATA);
					logWrite(hLog, ISYNCSTR_MSG_OP_INSERT);
					logWrite(hLog, ISYNCSTR_MSG_DATA);
					break;
				case ISCCONST_OpUpdate:
					PRINTF(isy_T("%s\n%s"), ISYNCSTR_MSG_OP_UPDATE, ISYNCSTR_MSG_DATA);
					logWrite(hLog, ISYNCSTR_MSG_OP_UPDATE);
					logWrite(hLog, ISYNCSTR_MSG_DATA);
					break;
				default:
					PRINTF(isy_T("%s\n%s"), ISYNCSTR_MSG_OP_INVALID, ISYNCSTR_MSG_DATA);
					logWrite(hLog, ISYNCSTR_MSG_OP_INVALID);
					logWrite(hLog, ISYNCSTR_MSG_DATA);
					break;
			}
		
			for (i=0; i<pConflict->colc; i++)
			{
				ISCLISTENCOLUMN *pColumn = &(pConflict->colv[i]);
		
				// PRINTF(isy_T(" (%i,%i,%i,%i) "), pColumn->pos, pColumn->pkseq, pColumn->type, pColumn->size);
				
				if (pColumn->data == NULL && pColumn->size == -1) {
					// no column data
				}
				else if (pColumn->data == NULL) {
					PRINTF (isy_T("%s"), ISYNCSTR_MSG_NULL);
					logWrite (hLog, ISYNCSTR_MSG_NULL);
		
				}
				else if (pColumn->type == SQL_BLOB) {
					PRINTF(isy_T(" [ * ] "));
					logWrite(hLog, isy_T(" [ * ] "));
				}
				else {
					isy_BYTE* pData;
		
					// for null terminated character
					pData = malloc(pColumn->size + 2);
		
					if (pData == NULL) break;
					
					memcpy(pData, &(pColumn->data[0]), pColumn->size);
		
					pData[pColumn->size] = 0;
		
					PRINTF(isy_T(" [ %s ]"), pData);					
					logWrite(hLog, isy_T(" [ "));
					logWrite(hLog, pData);
					logWrite(hLog, isy_T(" ]"));
		
					free(pData);
				}
			}
			PRINTF (isy_T(" ] \n"));
			logWrite (hLog, isy_T(" ] \n"));
		
			break;
		}
		case ISCEVTTYPE_Fatal:
		case ISCEVTTYPE_Error:
		default:
			printEventState(event, statusMsg);
			logEvent(hLog, event, statusMsg);
			PRINTF(isy_T("\n"));
			break;
	} // switch (event->type)
			
	return rc;
}

int getPrefs(Prefs* pPrefs)
{
	isy_TCHAR fileName[100];
	FILE* file;
	int fileSize = sizeof(Prefs);
	
	TSTR_CPY(fileName, isy_T(PROPFILE));
	file = FOPEN(fileName, isy_T("rb"));

	if (file == NULL)
		return -1;
	
	FREAD(pPrefs, fileSize, 1,  file);
	FCLOSE(file);

	if (TSTR_LEN(pPrefs->URL) == 0) {
		isy_TCHAR *protocol = TSTR_STR(pPrefs->dest, isy_T("://"));
		if (protocol == NULL) {
			// make http as the default transport protocol
			TSTR_CPY(pPrefs->URL, (pPrefs->options & ISYNCOPTION_SSL) ? isy_T("https://") : isy_T("http://"));
			TSTR_CAT(pPrefs->URL, pPrefs->dest);
			TSTR_N_CPY(pPrefs->dest, pPrefs->URL, 50);
			pPrefs->dest[49] = (isy_TCHAR) '\0';
		} else {
			TSTR_CPY(pPrefs->URL, pPrefs->dest);
		}
	}
	return 0;
}


void setPrefs(Prefs* pPrefs)
{
	isy_TCHAR fileName[50];
	FILE* file;

	TSTR_CPY(fileName, isy_T(PROPFILE));
	file = FOPEN(fileName, isy_T("wb+"));
	FWRITE(pPrefs, sizeof(Prefs), 1, file);
	FCLOSE(file);
}
	
/********************************************************************/

isy_TCHAR* appEventToMessage(ISCEVT*  event)
{	

	switch (event->code) 
	{
	case ISCEVT_InfSyncStarted:
		return ISYNCSTR_InfSyncStarted;
	case ISCEVT_InfPrepMsg:
		return ISYNCSTR_InfPrepMsg;
	case ISCEVT_InfSendMsg:
		return ISYNCSTR_InfSendMsg;
	case ISCEVT_InfWaitMsg:
		return ISYNCSTR_InfWaitMsg;
	case ISCEVT_InfApplyMsg:
		return ISYNCSTR_InfApplyMsg;
	case ISCEVT_InfCancelingSync:
		return ISYNCSTR_InfCancelingSync;
	case ISCEVT_InfSubsSetStarted:
		return ISYNCSTR_InfSubsSetStarted;
	case ISCEVT_InfSyncingSubs:
		return ISYNCSTR_InfSyncingSubs;
	case ISCEVT_InfSubsSetFailed:
		return ISYNCSTR_InfSubsSetFailed;
	case ISCEVT_InfSubsSetCanceled:
		return ISYNCSTR_InfSubsSetCanceled;
	case ISCEVT_InfSubsSetSucceeded:
		return ISYNCSTR_InfSubsSetSucceeded;
	case ISCEVT_InfSyncSucceeded:
		return ISYNCSTR_InfSyncSucceeded;
	case ISCEVT_InfSyncFailed:
		return ISYNCSTR_InfSyncFailed;
	case ISCEVT_InfSyncCanceled:
		return ISYNCSTR_InfSyncCanceled;
	case ISCEVT_InfSyncProg:
	case ISCEVT_InfSyncMsgSent:
		return ISYNCSTR_InfSyncProg;
	case ISCEVT_InfNoNewChange:
		return ISYNCSTR_InfNoNewChange;

	// Conflict events
	case ISCEVT_CftReject:
		return ISYNCSTR_CftReject;

	// Error events
	case ISCEVT_ErrOpenAdapter:
		return ISYNCSTR_ErrOpenAdapter;
	case ISCEVT_ErrLoadAdapter:
		return ISYNCSTR_ErrLoadAdapter;
	case ISCEVT_ErrCloseAdapter:
		return ISYNCSTR_ErrCloseAdapter;
	case ISCEVT_ErrAuthenticateKey:
		return ISYNCSTR_ErrAuthenticateKey;
	case ISCEVT_ErrClientCryptoFailed:
		return ISYNCSTR_ErrClientCryptoFailed;
	case ISCEVT_ErrEncryptNotAvail:
		return ISYNCSTR_ErrEncryptNotAvail;
	case ISCEVT_ErrEncryptLibOpen:
		return ISYNCSTR_ErrEncryptLibOpen;
	case ISCEVT_ErrSubsNotFound:
		return ISYNCSTR_ErrSubsNotFound;
	case ISCEVT_ErrSubsNotAvail:
		return ISYNCSTR_ErrSubsNotAvail;
	case ISCEVT_ErrSubsDefAltered:
		return ISYNCSTR_ErrSubsDefAltered;
	
	case ISCEVT_ErrCloseNetLib:
		return ISYNCSTR_ErrCloseNetLib;
	case ISCEVT_ErrOutOfMemory:
		return ISYNCSTR_ErrOutOfMemory;

	case ISCEVT_ErrAllocResource:
		return ISYNCSTR_ErrAllocResource;
	case ISCEVT_ErrConnectData:
		return ISYNCSTR_ErrConnectData;
	case ISCEVT_ErrDisconnectData:
		return ISYNCSTR_ErrDisconnectData;
	case ISCEVT_ErrNoData:
		return ISYNCSTR_ErrNoData;
	case ISCEVT_ErrSyncDisabled:
		return ISYNCSTR_ErrSyncDisabled;
	case ISCEVT_ErrServerException:
		return ISYNCSTR_ErrServerException;
	case ISCEVT_ErrMessageFormat:
		return ISYNCSTR_ErrMessageFormat;
	case ISCEVT_ErrNotFound:
		return ISYNCSTR_ErrNotFound;
	case ISCEVT_ErrEndOfData:
		return ISYNCSTR_ErrEndOfData;
	case ISCEVT_ErrDataTooLong:
		return ISYNCSTR_ErrDataTooLong;
	case ISCEVT_ErrReadOnly:
		return ISYNCSTR_ErrReadOnly;
	case ISCEVT_ErrOperation:
		return ISYNCSTR_ErrOperation;
	case ISCEVT_ErrUnauthorized:
		return ISYNCSTR_ErrUnauthorized;
	case ISCEVT_ErrNotAvailable:
		return ISYNCSTR_ErrNotAvailable;
	case ISCEVT_ErrNotSupported:
		return ISYNCSTR_ErrNotSupported;

	case ISCEVT_ErrInternal:
		return ISYNCSTR_ErrInternal;

	// Fatal events
	case ISCEVT_FatSyncCfgAbort:
		return ISYNCSTR_FatSyncCfgAbort;
	case ISCEVT_FatAuthenticateFailed:
		return ISYNCSTR_FatAuthenticateFailed;
	case ISCEVT_FatIncompVersion:
		return ISYNCSTR_FatIncompVersion;
	case ISCEVT_FatInvalidSession:
		return ISYNCSTR_FatInvalidSession;
	case ISCEVT_FatSyncGroup:
		return ISYNCSTR_FatSyncGroup;
	case ISCEVT_FatRegisterDevice:
		return ISYNCSTR_FatRegisterDevice;
	case ISCEVT_FatServerForbidden:
		return ISYNCSTR_FatServerForbidden;
	case ISCEVT_FatServerNotFound:
		return ISYNCSTR_FatServerNotFound;
	case ISCEVT_FatServerNotAvail:
		return ISYNCSTR_FatServerNotAvail;
	case ISCEVT_FatServer:
		return ISYNCSTR_FatServer;
	case ISCEVT_FatProtocolNotSupported:
		return ISYNCSTR_FatProtocolNotSupported;
	case ISCEVT_FatServerBusy:
		return ISYNCSTR_FatServerBusy;
	case ISCEVT_FatNetOpenConn:
		return ISYNCSTR_FatNetOpenConn;
	case ISCEVT_FatOpenNetLib:
		return ISYNCSTR_FatOpenNetLib;
	case ISCEVT_FatNetConnect:
		return ISYNCSTR_FatNetConnect;
	case ISCEVT_FatNetSend:
		return ISYNCSTR_FatNetSend;
	case ISCEVT_FatNetReceive:
		return ISYNCSTR_FatNetReceive;
	case ISCEVT_FatNetTimeout:
		return ISYNCSTR_FatNetTimeout;
	case ISCEVT_FatResolveHost:
		return ISYNCSTR_FatResolveHost;
	case ISCEVT_FatCreateImage:
		return ISYNCSTR_FatCreateImage;
	case ISCEVT_FatNetUnknown:
		return ISYNCSTR_FatNetUnknown;

	// Retry events
	case ISCEVT_TryServerBusy:
		return ISYNCSTR_TryServerBusy; 
	case ISCEVT_TryNetConnect:
		return ISYNCSTR_TryNetConnect; 
	case ISCEVT_TryNetSend:
		return ISYNCSTR_TryNetSend;
	case ISCEVT_TryNetReceive:
		return ISYNCSTR_TryNetReceive;
	case ISCEVT_TryNetTimeout:
		return ISYNCSTR_TryNetTimeout ;
	
	// Query events
	case ISCEVT_QueCancel:
		return ISYNCSTR_QueCancel;
	case ISCEVT_QueLogin:
		return ISYNCSTR_QueLogin;
	case ISCEVT_QueCancelUponError:
		return ISYNCSTR_QueCancelUponError;
	case ISCEVT_QueSubsTarget:
		return ISYNCSTR_QueSubsTarget;
	
	default:
		return isy_T("");
	}

	return NULL;
} // appEventToMessage

/********************************************************************/

static void openLog()
{
	isy_TCHAR filePath[128];
	
	if(TSTR_LEN(myPref.path) > 0)
	{
	    TSTR_CPY(filePath, myPref.path);
	    if(filePath[TSTR_LEN(filePath) - 1] != FPATH_SLASH)
		    TSTR_CAT(filePath, isy_T(FPATH_SLASHSTR));
	}
	else
		TSTR_CPY(filePath, isy_T(""));

	TSTR_CAT(filePath, isy_T(LOGFILE));
	hLog = FOPEN(filePath, isy_T("rb"));
}

static void rewindLog()
{
	isy_TCHAR filePath[128];
	isy_TCHAR buf[128];

	if(TSTR_LEN(myPref.path) > 0)
	{
	    TSTR_CPY(filePath, myPref.path);
	    if(filePath[TSTR_LEN(filePath) - 1] != FPATH_SLASH)
		    TSTR_CAT(filePath, isy_T(FPATH_SLASHSTR));
	}
	else
		TSTR_CPY(filePath, isy_T(""));

	TSTR_CAT(filePath, isy_T(LOGFILE));
	hLog = FOPEN(filePath, isy_T("wb"));
	logCodePage(hLog);

	if (hLog) {
		unsigned long version;
		int      verMajor, verMinor, verMod, verFix;
		isy_TCHAR     verStr[16];

		version = iscGetVersion();
		verMajor = (int) (version >> 24);
		verMinor = (int) ((version >> 16) & 0x000000FF);
		verMod = (int) ((version >> 8) & 0x000000FF);
		verFix = (int) ((version) & 0x000000FF);

		TSTR_ITOA(verStr, verMajor);
		TSTR_CAT(verStr, isy_T("."));
		TSTR_ITOA(verStr+TSTR_LEN(verStr), verMinor);
		TSTR_CAT(verStr, isy_T("."));
		TSTR_ITOA(verStr+TSTR_LEN(verStr), verMod);
		TSTR_CAT(verStr, isy_T("."));
		TSTR_ITOA(verStr+TSTR_LEN(verStr), verFix);

		appCurrentDateTime(buf, ISYNCSTR_DateTime);
		TSTR_CAT(buf, isy_T(" "));
		TSTR_CAT(buf, isy_T("DSYC000I:\tIBM SyncClient V."));
		TSTR_CAT(buf, verStr);
		TSTR_CAT(buf, FILE_NL);
		logWrite(hLog, buf);
		
	}
}

static void closeLog()
{
	if (hLog) {
		fclose(hLog);
		hLog = NULL;
	}
}


static void logEvent(FILE *hLog, ISCEVT*  event, isy_TCHAR* msg)
{
	isy_TCHAR buf[LOG_LINE_LEN];
	isy_TCHAR buf2[32];	

	// Get the timestamp
	appCurrentDateTime(buf, ISYNCSTR_DateTime);
	TSTR_CAT(buf, isy_T(" "));
	// Get the error/status code	
		
	TSTR_CAT(buf, isy_T("DSYC"));
	TSTR_ITOA(buf2, event->code);
	TSTR_CAT(buf, buf2);
	TSTR_CAT(buf, 
			 (event->type == ISCEVTTYPE_Info)     ? isy_T("I:\t") : \
			 (event->type == ISCEVTTYPE_Retry)    ? isy_T("R:\t") : \
			 (event->type == ISCEVTTYPE_Query)    ? isy_T("Q:\t") : \
			 (event->type == ISCEVTTYPE_Error)    ? isy_T("E:\t") : \
			 (event->type == ISCEVTTYPE_Fatal)    ? isy_T("F:\t") : \
			 (event->type == ISCEVTTYPE_Conflict) ? isy_T("C:\t") : isy_T("?"));
	
	TSTR_N_CAT(buf, msg, LOG_LINE_LEN - TSTR_LEN(buf) - 2);
	TSTR_CAT(buf, FILE_NL);

	logWrite(hLog, buf);
} // writeLog

static void logWrite(FILE *hLog, isy_TCHAR *buf)
{	
	FPRINTF(hLog, isy_T("%s"), buf);
}

const isy_TCHAR* appCurrentDateTime(isy_TCHAR* buf, isy_TCHAR* formatString)
{
	const isy_TCHAR* psz;
	int y, m, d, hr, mi, se;
	int i, k;
		
	psz = ISYNCSTR_DateTime;
	
	OSGET_CURRENT_DATE_TIME(&y, &m, &d, &hr, &mi, &se);
	for (i = 0, k = 0; psz[i]; i++)
	{
		switch (psz[i]) {
		case 'y':
			TSTR_ITOA(buf + k, y);
			k = TSTR_LEN(buf);
			break;
		case 'm':
			TSTR_ITOA(buf + k, m);
			k = TSTR_LEN(buf);
			break;
		case 'd':
			TSTR_ITOA(buf + k, d);
			k = TSTR_LEN(buf);
			break;
		case 'h':
			if (hr < 10) buf[k++] = '0';
			TSTR_ITOA(buf + k, hr);
			k = TSTR_LEN(buf);
			break;
		case 't':
			if (mi < 10) buf[k++] = '0';
			TSTR_ITOA(buf + k, mi);
			k = TSTR_LEN(buf);
			break;
		case 's':
			if (se < 10) buf[k++] = '0';
			TSTR_ITOA(buf + k, se);
			k = TSTR_LEN(buf);
			break;
		default:
			buf[k] = psz[i];
			++k;
		}
	}
	buf[k] = 0;
	
	return buf;
}

static void logCodePage(FILE* hLog)
{
/*
Byte-order mark		Description 
EF BB BF			UTF-8 
FE FF				UTF-16/UCS-2, little endian 
FF FE				UTF-16/UCS-2, big endian 
FF FE 00 00			UTF-32/UCS-4, little endian. 
00 00 FE FF			UTF-32/UCS-4, big-endian. 
*/
#ifdef _UNICODE
	FWRITE("\xFF", 1, 1, hLog);
	FWRITE("\xFE", 1, 1, hLog);
	FFLUSH(hLog);
#endif
}
