/************************************************************************
 *
 * Licensed Materials - Property of IBM
 * Program Name: IBM DB2Everyplace Sync Client GUI
 * (c ) Copyright IBM Corp. 1999-2003. All Rights Reserved.
 *
 * US Government Users Restricted Rights 
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. Licensed Materials - Property of IBM
 *
 * IBM grants you ("Licensee") a non-exclusive, royalty free, license to
 * use Sample Code.  This software is provided as-is and without any
 * warranties, whether EXPRESS OR  IMPLIED, INCLUDING ANY IMPLIED
 * WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR 
 * NON-INFRINGEMENT. IBM AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY 
 * DAMAGES SUFFERED BY LICENSEE THAT RESULT FROM YOUR USE OF THE SOFTWARE.
 * IN NO EVENT WILL IBM OR ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE,
 * PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, 
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE 
 * THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE 
 * SOFTWARE, EVEN IF IBM HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES.
 *
 ************************************************************************
 */
/************************************************************************
 *
 * File Name:	ISyncSample.java
 * Description:	Sample application to demonstrate the ISync Java Client API
 *
 ************************************************************************
 */

import java.lang.*;
import java.sql.*;
import java.util.*;

/*
	Import the ISync Java Client Package
*/
import com.ibm.mobileservices.isync.*; 
import com.ibm.mobileservices.isync.event.*; 

/**
	<p>
	This is a sample application to demonstrate the ISync Java Client API.
	</p>

	<p>
	Usage: java ISyncSample &lt;property file&gt;
	e.g java ISyncSample isyncdb2e
	</p>	

	<p>
	You must specify a .properties file specific to each provider that contains 
	the synchronization parameters.
		DB2:	isyncdb2
		DB2e:	isyncdb2e
		DB2j:	isyncdb2j
	</p>
	
	<p>	
	Required ISync libraries:
		DB2/DB2j: 	db2jisync.jar
		DB2e:		isync4j.jar + native libraries
	</p>
*/

public class ISyncSample implements ISyncListener
{
	// Synchronization objects
	private ISyncProvider provider;
	private ISyncService service;
	private ISyncConfigStore config;
	private ISyncDriver syncer;
	
	// Synchronization parameters

	private static PropertyResourceBundle configProps;
	private Properties userProps = new Properties();
	private ISyncSubscriptionSet ssArr[] = null;

	/*
		Implement the eventIssued() method in the ISyncListener
		interface if you are interested in event notification (optional)
	*/

	public int eventIssued(ISyncEvent evt) 
	{
		int evtType = evt.getEventType();
		int evtCode = evt.getEventCode();
		switch(evtType) 
		{
			// Display event status
			case ISync.EVTTYPE_INFO: 
			case ISync.EVTTYPE_ERROR:

				System.out.println ("*********************");
				System.out.println ("SubsSet:    " + evt.getSubscriptionSetName());
				System.out.println ("Subs:       " + evt.getSubscriptionName());
				System.out.println ("SubsType:   " + evt.getSubscriptionType());
				System.out.println ("Event Type: " + evtType);
				System.out.println ("Event Code: " + evt.getEventCode());
				System.out.println ("Progress:   " + evt.getSyncProgress());
				System.out.println ("**********************\n");
				return ISync.RTNCB_DONE;

			case ISync.EVTTYPE_RETRY:

				System.out.println ("*********************");
				System.out.println("Retry...");
				System.out.println ("*********************");
				return ISync.RTNCB_REPLY_YES;

			case ISync.EVTTYPE_CONFLICT:

				if (evtCode == ISync.EVT_CFT_REJECT) 
				{
					System.out.println ("*********************");
					System.out.println("Conflict detected!");
					System.out.println ("*********************");

					ConflictReader cr = (ConflictReader)evt.getEventInfo();
					processRejectedData(cr);
					return ISync.RTNCB_DONE;
				}

				break;

			case ISync.EVTTYPE_QUERY:

				// Local data encryption case - DB2e only (optional)
				if (evtCode == ISync.EVT_QUE_LOGIN)
				{
					String login[] = (String [])evt.getEventInfo();
					System.out.println("Querying login for local encryption.");
					System.out.println("Target: " + login[0]);

					login[1] = userProps.getProperty("target.db.user");
					login[2] = userProps.getProperty("target.db.password");
					return ISync.RTNCB_DONE;
				}
				else if (evtCode == ISync.EVT_QUE_SUBSCRIPTION_TARGET)
				{
					// Specify multiple directories for sync (optional)
					int subsType = evt.getSubscriptionType();
					String login[] = (String [])evt.getEventInfo();

					if (subsType == ISync.SUBSTYPE_FILE)
						login[0] = userProps.getProperty("filesubdir");
					else
						login[0] = userProps.getProperty("jdbcsuburl");						

					System.out.println("Sample set subs " + login[0]);
					return ISync.RTNCB_DONE;
				}
				
				break;

			// Ignore other event types
			default:
				break;
		}

		// Let sync engine take default action
		return ISync.RTNCB_DEFAULT;
	}

	// Sample processes INSERT and UPDATE operations only
	private void processRejectedData(ConflictReader cr) 
	{
		int op;
		ResultSet pkrs = null;
		ResultSet rs = null;
		String cmsg;

		if (cr == null) 
			return;

		try
		{ 
			op = cr.getOperation();
			cmsg = "Conflicts detected in Table: " + cr.getTableName();
			System.out.println (cmsg);
			if (op != ISync.OP_DELETE)
			{
				rs = cr.getRejectedRows();
				processResultSet(rs);
			}
			else
				System.out.println ("A deleted row was rejected");
		}
		catch (ISyncException iex) 
		{
			iex.printStackTrace();
			System.out.println (iex.getCode());
		}
		catch (SQLException sqlex)
		{
			sqlex.printStackTrace();

			while(sqlex != null)
			{
				System.out.println("SQLERROR: \n" + sqlex.getErrorCode() + 
					", SQLState: " + sqlex.getSQLState() + 
					", Message: " + sqlex.getMessage() +
					", Vendor: " + sqlex.getErrorCode());
				sqlex = sqlex.getNextException();
			}
		}

		System.out.println("");
	} // processRejectedData


	private void processResultSet(ResultSet rs) throws SQLException 
	{
		ResultSetMetaData rsmd = rs.getMetaData();
		int numCols = rsmd.getColumnCount();
		String name;
		int type;
		int i = 0;

		while (rs.next()) 
		{
			i = 1;

			while (i <= numCols) 
			{
				name = rsmd.getColumnName(i);
				type = rsmd.getColumnType(i);
				System.out.print (name + ":");

				//BLOB - use getBytes()
				if (type == Types.LONGVARBINARY)
				{
					byte[] data = rs.getBytes(i);

					if (data == null) 
					{
						System.out.print(" ( " + type + " ) ");
						System.out.print(" [ NULL ] ");
					}
					else 
					{
						System.out.print(" ( " + type + " ) ");
						System.out.println(" [ BLOB:" + data.length + " ] ");
					}
				}
				//Sample uses getString() for other types
				else 
				{
					String s = rs.getString(i);

					if (s == null) 
					{
						System.out.print(" ( " + type + " ) ");
						System.out.print(" [ NULL ] ");
					}
					else 
					{
						System.out.print(" ( " + type + " ) ");
						System.out.print(" [ " + s + " ] ");
					}
				}

				i++;

			} //while numCols

		} //while rs.next

		rs.close();

	} //processResultSet

	private void readProperties(String propFile)
	{
		// Required properties for sample
		try
		{
			configProps = (PropertyResourceBundle) ResourceBundle.getBundle(propFile);
			
			userProps.put("syncdriver", configProps.getString("syncdriver"));			
			userProps.put("syncprotocol", configProps.getString("syncprotocol"));
			userProps.put("server.url", configProps.getString("server.url"));
			userProps.put("target.db.driver", configProps.getString("target.db.driver"));
			userProps.put("target.db.url", configProps.getString("target.db.url"));
			userProps.put("isync.user", configProps.getString("isync.user"));
			userProps.put("isync.password", configProps.getString("isync.password"));
			userProps.put("isync.timeout", configProps.getString("isync.timeout"));
			userProps.put("isync.trace", configProps.getString("isync.trace"));
			userProps.put("path", configProps.getString("path"));
			userProps.put("jdbcsuburl", configProps.getString("jdbcsuburl"));
			userProps.put("filesubdir", configProps.getString("filesubdir"));												
		}
		catch (MissingResourceException mre)
		{
			throw mre;
		}
		
		// Optional properties for sample
		try
		{
			if (configProps.getString("target.db.user") != null)
				userProps.put("target.db.user", configProps.getString("target.db.user"));
			if (configProps.getString("target.db.password") != null)
				userProps.put("target.db.password", configProps.getString("target.db.password"));
		}
		catch (MissingResourceException mre)
		{
		}
		
	} //readProperties

	public void runSample(String propFile)
	{
		try 
		{
			int rc = 0;

			// Read properties file
			readProperties(propFile);
			
			// Load sync driver
			Class.forName(userProps.getProperty("syncdriver"));

			// Get an instance of synchronization service from the manager

			provider = ISyncManager.getISyncProvider(userProps.getProperty("syncprotocol"));
			service = provider.createSyncService(userProps.getProperty("server.url"), userProps);

			// Get an instance of the configuration store

			config = service.getConfigStore(userProps.getProperty("path"));

			// Get an instance of the sync driver to perform synchronization

			syncer = config.getSyncDriver(); 

			// Set the listener object for event notification from the syncer object 
			// during synchronization (optional)

			syncer.setSyncListener(this);

			// Perform synchronization on all enabled subscription sets
            
			rc = syncer.sync();

			switch (rc) 
			{
				case ISync.RTN_SUCCEEDED:
					System.out.println("Synchronization succeeded");
					break;

				case ISync.RTN_CANCELED:
					System.out.println ("Synchronization canceled");
					break;

				default: 
					System.out.println ("Synchronization failed");
					break;
			}

			ssArr = config.getSubscriptionSets();

			for (int i=0; i < ssArr.length; i++) 
			{

				System.out.print ("Subscription Set: " + ssArr[i].getName() + " Status: ");

				switch(ssArr[i].getStatus())
				{
					case ISync.STATUS_READY:
						System.out.println("READY");
						break;

					case ISync.STATUS_COMPLETED:
						System.out.println ("COMPLETED");
						break;

					case ISync.STATUS_CANCELED: 
						System.out.println ("CANCELED");
						break;

					default: 
						System.out.println ("FAILED");
						break;
				}
			}
		}
		catch (ISyncException ie)
		{
			System.out.println("Exception code: " + ie.getCode());
			ie.printStackTrace();
		}
		catch (Exception e) 
		{
			e.printStackTrace();
		}      
		finally 
		{
			try 
			{
				// Close and free allocated resources by the syncer object
				if (syncer != null)
				{
					syncer.close();
					syncer = null;
				}

				// Close and free allocated resources by the config object
				if (config != null) 
				{
					config.close();
					config = null;
				}

				// Close and free allocated resources by the service object
				if (service != null) 
				{
					service.close();
					service = null;
				}
			}
			catch(ISyncException ie2)
			{
				System.out.println("Exception code: " + ie2.getCode());
				ie2.printStackTrace();
			}
		} //finally
	} //runSample

	public static void main (String[] args) 
	{ 
		if (args.length == 1) 
		{
			ISyncSample sample = new ISyncSample();
			sample.runSample(args[0]);
		}
		else 
			System.out.println("Usage: java ISyncSample property_file");
	}

} //class ISyncSample
