/************************************************************************
 *
 * Licensed Materials - Property of IBM
 * Program Name: IBM DB2Everyplace Sync Client
 * (c ) Copyright IBM Corp. 1999-2003. All Rights Reserved.
 *
 * US Government Users Restricted Rights 
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. Licensed Materials - Property of IBM
 *
 * IBM grants you ("Licensee") a non-exclusive, royalty free, license to
 * use Sample Code.  This software is provided as-is and without any
 * warranties, whether EXPRESS OR  IMPLIED, INCLUDING ANY IMPLIED
 * WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR 
 * NON-INFRINGEMENT. IBM AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY 
 * DAMAGES SUFFERED BY LICENSEE THAT RESULT FROM YOUR USE OF THE SOFTWARE.
 * IN NO EVENT WILL IBM OR ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE,
 * PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, 
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE 
 * THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE 
 * SOFTWARE, EVEN IF IBM HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES.
 *
 ************************************************************************
 */

using System;
using System.Collections;
using System.IO;
using System.Xml;

using IBM.Data.Sync;

#if (!UNDER_CF)
using IBM.Data.Sync.DB2e;		//standard framework
#else
using IBM.Data.Sync.DB2e.CF;	//compact framework
#endif

namespace DB2Sync
{
	/// <summary>
	/// Command interface where a series of prompts
	/// allow control of synchronization
	/// </summary>
	/// <remarks>
	/// Usage: DB2SyncConsole.exe
	/// <p>
	/// ISync managed provider and sync client native libraries 
	/// must be included in environment path
	/// </p>
	/// </remarks>
	public class DB2SyncConsole : DB2SyncConstants 
	{
		ISyncProvider provider;
		ISyncService service;
		ISyncConfigStore config;
		ISyncDriver syncer;

		Hashtable isyncProps;
		FileStream hLog;
		StreamReader logfileReader;
		StreamWriter logfileWriter;

		/// <summary>
		/// Constructor
		/// </summary>
		public DB2SyncConsole() 
		{
			isyncProps = new Hashtable();
		}

		/// <summary>
		/// Loops until GetAction switches to DO_EXIT and the
		/// program is terminated
		/// </summary>
		public void Run()
		{
			int rc;
			String action;

			rc = InitSettings();
			
			if (rc < 0)
			{
				Console.WriteLine("Invalid configuration file: "+ PROPFILE);
				logfileWriter.Write("Invalid configuration file: "+ PROPFILE);
			}
			else
			{
				Console.WriteLine("\n\n");
				while (true) 
				{
					try
					{
						action = GetAction();
				
						try
						{
							if (action == null)
								continue;
							int i = Int32.Parse(action);
						}
						catch (FormatException e) 
						{
							Console.WriteLine(e.Message);
							continue;
						}

						if (action.Equals(DO_SYNC)) 
						{
							Console.Write("\n\n");
							GoSync();
						} 
						else if (action.Equals(DO_APPS)) 
						{
							Console.Write("\n\n");
							GoApps();
						} 
						else if (action.Equals(DO_PREFS)) 
						{
							Console.Write("\n\n");
							GoSettings();
						} 
						else if (action.Equals(DO_LOG)) 
						{
							Console.Write("\n\n");
							GoLog();
						} 
						else if (action.Equals(DO_ABOUT)) 
						{
							GoAbout();
						} 
						else if (action.Equals(DO_EXIT)) 
						{
							break;
						} 
						else if (action.Equals(DO_NOTHING)) 
						{
						} 
						else 
						{
							Console.WriteLine(action + " " + DB2SYNC_InvalidChoice);
						}
					}
					catch (ISyncException ise)
					{
						Console.WriteLine(ise);
					}

				} // while loop
			} // if InitSettings

			GoExit();
		} // startClient

		/// <summary>
		/// Prompts the user for an action to perform and returns the choice
		/// </summary>
		/// <returns>choice as string</returns>
		public String GetAction()
		{
			String action = DO_NOTHING;

			Console.WriteLine(DB2SYNC_MainMenu);
			Console.WriteLine(DB2SYNC_MainMenu1);
			Console.WriteLine(DB2SYNC_MainMenu2);
			Console.WriteLine(DB2SYNC_MainMenu3);
			Console.WriteLine(DB2SYNC_MainMenu4);
			Console.WriteLine(DB2SYNC_MainMenu5);
			Console.WriteLine(DB2SYNC_MainMenu6);
			Console.Write(DB2SYNC_MakeChoice);

			action = Console.ReadLine();
			return action.Trim();
		}

		/// <summary>
		/// Performs the synchronization.  Before Sync() is called, 
		/// SyncEvent is set so that the application can receive callbacks
		/// </summary>
		public void GoSync()
		{
			int rc;
			DB2SyncListener iSyncListener = new DB2SyncListener(isyncProps);

			rc = SyncStartup();
			if (rc != ISync.RTN_SUCCEEDED) 
			{
				Console.Write(
					"\nSync Engine startup error. Make sure path is valid. Error code: "
					+ rc
					+ "\n\n\n");
				return;
			}

			RewindLog();		
			iSyncListener.SetOutPutStream(logfileWriter);		

			syncer.SyncEvent += new ISyncEventHandler(iSyncListener.EventIssued);

			rc = syncer.Sync();

			Console.Write("iscEngineSync() = " + rc + "\n\n");

			ISyncSubscriptionSet []ssArr = config.GetSubscriptionSets();
			if (ssArr.Length == 0) 
			{
				Console.WriteLine(DB2SYNC_NoSubsetInfo);
			}
			else
			{
				//Print status for each subscription set
				for (int i = 0; i < ssArr.Length; i++) 
				{
					Console.WriteLine(DB2SYNC_SyncStatus);

					if (ssArr[i].IsEnabled()) 
					{
						Console.Write("+   ");
					} 
					else 
					{
						Console.Write("-   ");
					}

					int subStatus = ssArr[i].GetStatus();

					if (subStatus == ISync.RTN_CANCELED)
						Console.Write("Cancelled");
					else if (subStatus == ISync.RTN_FAILED)
						Console.Write("Failed");
					else if (subStatus == ISync.RTN_NOT_FOUND)
						Console.Write("Not Found");
					else if (subStatus == ISync.RTN_READY)
						Console.Write("Ready");
					else if (subStatus == ISync.RTN_SUCCEEDED)
						Console.Write("Succeeded!");
					else
						Console.Write("" + subStatus);

					Console.Write(
						"   " + ssArr[i].GetId() + "   " + ssArr[i].GetName());

					Console.Write("\n");
					Console.Write(DB2SYNC_Line);

					Console.Write("\n\n");
				}
			}

			SyncShutdown();

		} // GoSync

		/// <summary>
		/// Allows the user to enable, disable, or reset a subscription
		/// set. If no changes are necessary to the sync application the user
		/// can press enter to accept the default setting.  
		/// </summary>
		public void GoApps()
		{
			ISyncSubscriptionSet []ssArr = null;
			int rc = 0;

			rc = SyncStartup();
			if (rc != ISync.RTN_SUCCEEDED) 
			{
				Console.Write(
					"\nSync Engine startup error. Make sure path is valid. Error code: "
					+ rc
					+ "\n\n\n");
				return;
			}

			Console.WriteLine(DB2SYNC_SyncApps);

			ssArr = config.GetSubscriptionSets();
			if (ssArr.Length == 0) 
			{
				Console.WriteLine(DB2SYNC_NoSubsetInfo+"\n\n");
			} 
			else
			{
				//For each subscription set, show current enable/disable and
				//refresh/synchronize condition; prompt for any changes.
				for (int i = 0; i < ssArr.Length; i++) 
				{
					String input;

					Console.Write("Subscription set " + ssArr[i].GetName());
					if (ssArr[i].IsEnabled())
						Console.Write(" " + DB2SYNC_Enabled);
					else
						Console.Write(" " + DB2SYNC_Disabled);

					if (ssArr[i].IsReset())
						Console.Write(" " + DB2SYNC_ModeRefresh);
					else
						Console.Write(" " + DB2SYNC_ModeSynchronize);

					Console.WriteLine("\n" + DB2SYNC_SetAppMode);
					Console.WriteLine(DB2SYNC_SetAppModeA);
					Console.WriteLine(DB2SYNC_SetAppModeB);
					Console.WriteLine(DB2SYNC_SetAppModeC);
					Console.WriteLine(DB2SYNC_SetAppModeD);
					Console.Write(DB2SYNC_MakeChoice);

					input = Console.ReadLine();
					input = input.Trim();

					if (input.Equals(MODE_ENABLE)) 
					{
						ssArr[i].Enable();
					} 
					else if (input.Equals(MODE_DISABLE)) 
					{
						ssArr[i].Disable();
					} 
					else if (input.Equals(MODE_RESET)) 
					{
						ssArr[i].Reset();
					} 
					else if (input.Equals(MODE_DONE)) 
					{
					}
					Console.Write("\n\n");
				}
			}

			SyncShutdown();
		} // GoApps

		/// <summary>
		/// Read application preferences.  Create default application 
		/// preferences if DB2SYNC.config does not exist
		/// </summary>
		/// <returns>0 upon success; non-zero otherwise</returns>
		public int InitSettings()
		{
			int rc = 0;
			if (File.Exists(PROPFILE))
				rc = GetPrefs(PROPFILE);
			else
			{
				isyncProps.Add("server", DEFAULT_SERVER);
				isyncProps.Add("port", DEFAULT_PORT);
				isyncProps.Add("isync.user", DEFAULT_USER);
				isyncProps.Add("isync.password", DEFAULT_PASSWORD);
				isyncProps.Add("isync.trace", DEFAULT_TRACE);
				isyncProps.Add("isync.timeout", DEFAULT_TIMEOUT);
				isyncProps.Add("path", DEFAULT_PATH);
				SetPrefs(PROPFILE);
			}

			OpenLog(false);
			return rc;
		} // InitSettings

		/// <summary>
		/// Allow the user to change the network and user settings which
		/// enable the sync connection to the server.  If no changes are 
		/// necessary, the user can press enter to accept the current setting.
		/// </summary>
		public void GoSettings()
		{
			String buf;
			int i, validTimeout;

			Console.WriteLine(DB2SYNC_Settings);
			Console.WriteLine(DB2SYNC_SetSettings);

			//Get hostname or ip
			Console.Write(DB2SYNC_Server + " [" + (String)isyncProps["server"] + "]: ");
			buf = Console.ReadLine();
			buf = buf.Trim();
			if (!buf.Equals(""))
				isyncProps["server"] = buf;

			//Get port
			Console.Write(DB2SYNC_Port + " [" + (String)isyncProps["port"] + "]: ");
			buf = Console.ReadLine();
			buf = buf.Trim();
			if (!buf.Equals(""))
				isyncProps["port"] = buf;

			//Get username	
			Console.Write(DB2SYNC_User + " [" + (String)isyncProps["isync.user"] + "]: ");
			buf = Console.ReadLine();
			buf = buf.Trim();
			if (!buf.Equals(""))
				isyncProps["isync.user"] = buf;

			//Get password
			Console.Write(DB2SYNC_Password + " [");
			for (i = ((String)isyncProps["isync.password"]).Length; i > 0; i--)
				Console.Write("*");
			Console.Write("]: ");
			buf = Console.ReadLine();
			buf = buf.Trim();
			if (!buf.Equals(""))
				isyncProps["isync.password"] = buf;

			//Get target directory path
			Console.Write(DB2SYNC_Path + " [" + (String)isyncProps["path"] + "]: ");
			buf = Console.ReadLine();
			buf = buf.Trim();
			if (!buf.Equals(""))
				isyncProps["path"] = buf;

			//Get trace preference
			Console.Write(
				DB2SYNC_Trace
				+ " ("
				+ DB2SYNC_Detailed
				+ "/"
				+ DB2SYNC_Default
				+ ")? ["
				+ (String)isyncProps["isync.trace"]
				+ "]: ");
			buf = Console.ReadLine();
			buf = buf.Trim();
			if (!buf.Equals(""))
				isyncProps["isync.trace"] = buf;

			while (true) 
			{
				//Get timeout length	
				validTimeout = Int32.Parse((String)isyncProps["isync.timeout"]);
				if (validTimeout == 0) 
				{
					Console.Write(
						DB2SYNC_Timeout + " [" + DB2SYNC_MinimumTimeout + "]: ");
				} 
				else if (validTimeout == -1) 
				{
					Console.Write(
						DB2SYNC_Timeout + " [" + DB2SYNC_NeverTimeout + "]: ");
				} 
				else 
				{
					validTimeout /= 60; // minutes
					Console.Write(
						DB2SYNC_Timeout + " [" + validTimeout + " min]: ");
				}

				buf = Console.ReadLine();
				buf = buf.Trim();
				try 
				{
					if (buf.Equals(""))
						break;

					validTimeout = Int32.Parse(buf);
					isyncProps["isync.timeout"] = ("" + validTimeout * 60);

					//valid timeout received. break.
					break;
				} 
				catch (FormatException e) 
				{
					Console.WriteLine(DB2SYNC_InvalidTimeout);
					Console.WriteLine(e);
					validTimeout = 0;
				}
			}

			Console.Write("\n\n");

			SetPrefs(PROPFILE);
		} // GoSettings

		/// <summary>
		/// Displays application log file on console
		/// </summary>
		public void GoLog()
		{
			try 
			{
				OpenLog(false);
				String buf;
				Console.WriteLine(DB2SYNC_LogFile);
				while ((buf = logfileReader.ReadLine()) != null) 
				{
					Console.WriteLine(buf);
				}
				Console.Write("\n\n");
			} 
			catch (FileNotFoundException fnf) 
			{
				Console.WriteLine(DB2SYNC_LogFile);
				Console.WriteLine(fnf);
				Console.WriteLine("\n\n");
			}
		} // GoLog

		/// <summary>
		/// Close synchronization objects
		/// </summary>
		public void SyncShutdown()
		{
			if (syncer != null)
			{
				syncer.Close();
				syncer = null;
			}

			if (config != null)
			{
				config.Close();
				config = null;
			}

			if (service != null)
			{
				service.Close();
				service = null;
			}
		}

		/// <summary>
		/// Create synchronization objects
		/// </summary>
		/// <returns>ISync.RTN_SUCCEEDED (zero) upon success; non-zero otherwise</returns>
		public int SyncStartup()
		{
			int rc = ISync.RTN_SUCCEEDED;

			SyncShutdown();

			//get an instance DB2eISyncProvider 
			provider = DB2eISyncProvider.GetInstance();

			//get an instance of synchronization service from the provider

			String uri = "http://" + (String)isyncProps["server"];
			if ((String)isyncProps["port"] != null)
				uri += ":" + (String)isyncProps["port"];

			service = provider.CreateSyncService(uri, isyncProps);

			//get an instance of the configuration store
			config = service.GetConfigStore((String)isyncProps["path"]);

			//get an instance of the sync driver to perform synchronization
			syncer = config.GetSyncDriver();

			return rc;
		}

		/// <summary>
		/// Displays client version information on console
		/// </summary>
		public void GoAbout()
		{
			String about;
			int rc;

			rc = SyncStartup();
			if (rc != ISync.RTN_SUCCEEDED) 
			{
				Console.Write(
					"\nSync Engine startup error. Make sure path is valid. Error code: "
					+ rc
					+ "\n\n\n");
				return;
			}
			about = syncer.ToString();
			Console.Write("\n" + DB2SYNC_About + "\n" + about + "\n\n");
			SyncShutdown();
		} // GoAbout

		/// <summary>
		/// Exits the program
		/// </summary>
		public void GoExit()
		{
			if (logfileWriter != null)
			{
				logfileWriter.Close();
				logfileWriter = null;
			}

			if (logfileReader != null)
			{
				logfileReader.Close();
				logfileReader = null;
			}

			SetPrefs(PROPFILE);
			SyncShutdown();
		} // GoExit

		/// <summary>
		/// Get application preferences.  Preference file DB2SYNC.config
		/// based on ConfigurationSettings.AppSettings format.
		/// </summary>
		/// <param name="fileName">name of file (DB2SYNC.config)</param>
		/// <returns>true if preference file is valid; false otherwise</returns>
		public int GetPrefs(String fileName)
		{
/*
			//DB2SyncConsole.exe.config gets deleted on build
			//Use DB2SyncConsole.config
*/
			XmlTextReader reader = new XmlTextReader(fileName);
			try
			{
				reader.WhitespaceHandling = WhitespaceHandling.None;

				while (reader.Read())
				{
					switch (reader.NodeType)
					{
						case XmlNodeType.Element:
						{
							if (reader.Name.Equals("add"))
							{
								if (!reader.MoveToFirstAttribute())
									break;
								String keyAttrib = reader.Name;
								String key = reader.Value;

								if (!reader.MoveToNextAttribute())
									break;
								String valueAttrib = reader.Name;
								String value = reader.Value;

								if (keyAttrib.Equals("key") && valueAttrib.Equals("value"))
									isyncProps.Add(key, value);
							}
							break;
						}
						default:
							break;
					} // switch
				} // while
			}
			finally
			{
				if (reader!=null)
					reader.Close();
			}

			if (!isyncProps.ContainsKey("server") || !isyncProps.ContainsKey("port")
				|| !isyncProps.ContainsKey("isync.user") || !isyncProps.ContainsKey("isync.password"))
				return -1;
			else
				return 0;
		} // GetPrefs

		/// <summary>
		/// Save application preferences.  Preference file DB2SYNC.config
		/// based on ConfigurationSettings.AppSettings format
		/// </summary>
		/// <param name="fileName">name of file (DB2SYNC.config)</param>
		public void SetPrefs(String fileName)
		{
			XmlTextWriter writer = new XmlTextWriter(fileName, System.Text.UTF8Encoding.UTF8);
			writer.Formatting = Formatting.Indented;
			writer.WriteStartDocument();
			writer.WriteStartElement("configuration");
			writer.WriteStartElement("appSettings");

			IDictionaryEnumerator e = isyncProps.GetEnumerator();
			while (e.MoveNext())
			{
				writer.WriteStartElement("add");
				writer.WriteAttributeString("key", (String) e.Key);
				writer.WriteAttributeString("value", (String) e.Value);
				writer.WriteEndElement();
			}

			writer.WriteEndElement();
			writer.WriteEndElement();

			writer.Close();
		} // SetPrefs

		/// <summary>
		/// Open application log in current directory
		/// </summary>
		/// <param name="bCreateNew">true to create new log</param>
		public void OpenLog(bool bCreateNew)
		{
			String filePath = "";
//use current directory
/*
			int pathLen = 0;

			if ((String)isyncProps["path"] != null)
				pathLen = ((String)isyncProps["path"]).Length;

			if (pathLen > 0) 
			{
				filePath += (String)isyncProps["path"];
				if (filePath[pathLen - 1] != Path.DirectorySeparatorChar)
					filePath += Path.DirectorySeparatorChar;
			}
*/
			filePath += LOGFILE;

			if (logfileWriter != null)
			{
				logfileWriter.Close();
				logfileWriter = null;
			}

			if (logfileReader != null)
			{
				logfileReader.Close();
				logfileReader = null;
			}

			if (bCreateNew)
				hLog = File.Open(filePath, FileMode.Create);
			else
				hLog = File.Open(filePath, FileMode.OpenOrCreate);
			logfileWriter = new StreamWriter(hLog);
			logfileReader = new StreamReader(hLog);
		} //OpenLog

		/// <summary>
		/// Calls OpenLog to create a new application log
		/// </summary>
		public void RewindLog()
		{
			OpenLog(true);

			if (logfileWriter != null)
			{
				long version;
				int verMajor, verMinor, verMod, verFix;

				String verStr = "";
				version = provider.GetVersion();
				verMajor = (int) (version >> 24);
				verMinor = (int) ((version >> 16) & 0x000000FF);
				verMod = (int) ((version >> 8) & 0x000000FF);
				verFix = (int) ((version) & 0x000000FF);

				verStr += verMajor;
				verStr += ".";
				verStr += verMinor;
				verStr += ".";
				verStr += verMod;
				verStr += ".";
				verStr += verFix;

				logfileWriter.WriteLine(DateTime.Now.ToShortDateString() + " " 
					+ DateTime.Now.ToLongTimeString() + " "
					+ "IBM SyncClient V." + verStr);
			}
		} // RewindLog

		/// <summary>
		/// DB2SyncConsole.exe entry point
		/// </summary>
		/// <param name="args">no arguments</param>
		public static void Main(string[] args)
		{
			DB2SyncConsole DB2Sync = new DB2SyncConsole();

			try 
			{
				DB2Sync.Run();
			}
			catch (Exception e) 
			{			
				Console.WriteLine(e);
			}
			finally
			{
				Console.Write(DB2SYNC_Exitting + " ...\n");
				Console.ReadLine();
			}
		} // Main

	} // class DB2SyncConsole

} // namespace DB2Sync
